function cse15(o)
% cse15 (cse15 companion code)
% by Christian Himpe, 2015 ( http://wwwmath.uni-muenster.de/u/himpe )
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )
%*

if(exist('emgr')~=2)
    disp('emgr framework is required. Download at http://gramian.de/emgr.m');
    return;
end

warning off all;

rand('seed',1009);
randn('seed',1009);

%% Setup

J = 1;
N = 64;
O = 8;
T = [0.0,0.01,1.0];
L = (T(3)-T(1))/T(2);
U = [ones(J,1),zeros(J,L-1)];
X = zeros(N,1);

A = rand(N,N);
P = A(:);

B = rand(N,J);
C = rand(O,N);

a = ones(N);
q = a(:);
Q = numel(q);

norm2 = @(x) sqrt(sum(sum(x.*x)));

A0 = -0.6*N*speye(N);

F = @(x,u,p) (A0 + reshape(p,[N,N]))*tanh(x) + B*u;
G = @(x,u,p) C*x;

%% Main

Y = irk3(F,G,T,X,U,P); % plot(T(2):T(2):T(3),Y); return;
Yd = Y + 0.01*max(Y(:))*randn(O,L);

r = 0.00001;

k2g = zeros(8,8);
l2g = zeros(8,8);

tic;
WJ = emgr(F,G,[J,N,O],T,'j',q,[0,0,0,0,0,0,0,0,0,1,0,0],1,0,1);
[UU,D,VV] = svd(WJ{1});
[PP,D,QQ] = svd(WJ{2});
OFFLINE_WJ = toc

a = 1;

for I=1:N/8:N
    uu = UU(:,1:I);
    vv = uu';

    b = 1;

    for K=1:N/8:N
        pp = PP(:,1:K);
        qq = pp';

        f = @(x,u,p) vv*F(uu*x,u,pp*p);
        g = @(x,u,p)    G(uu*x,u,pp*p);
        x = vv*X;

        j = @(p) sqrt(T(2))*norm2(Yd-irk3(f,g,T,x,U,p)) + r*(p'*p);
        p = fminsearch(j,qq*0.5*q,optimset('Display','off'));

        y = irk3(f,g,T,x,U,qq*P); % Reduced Order
        k2g(a,b) = norm2(Y-y)/norm2(Y);

        y = irk3(f,g,T,x,U,p); % Reduced Order
        l2g(a,b) = norm2(Y-y)/norm2(Y);

        b = b + 1;
        fprintf('#');
    end;

    a = a + 1;
    fprintf('\n');
end;

k2g
l2g

save('l2.mat','k2g','l2g');


k2o = zeros(8,8);
l2o = zeros(8,8);

tic;
XP = optmor(F,G,[J,N,O],T,N,0.5*q,[0,0,0,0,0,0,0,0.01,0.01,2,0,0],X,U,1,Yd);
UU = XP{1};
PP = XP{2};
OFFLINE_XP = toc

a = 1;

for I=1:N/8:N
    uu = UU(:,1:I);
    vv = uu';

    b = 1;

    for K=1:N/8:N
        pp = PP(:,1:K);
        qq = pp';

        f = @(x,u,p) vv*F(uu*x,u,pp*p);
        g = @(x,u,p)    G(uu*x,u,pp*p);
        x = vv*X;

        j = @(p) sqrt(T(2))*norm2(Yd-irk3(f,g,T,x,U,p)) + r*(p'*p);
        p = fminunc(j,qq*0.5*q,optimset('Display','off'));

        y = irk3(f,g,T,x,U,qq*P); % Reduced Order
        k2o(a,b) = norm2(Y-y)/norm2(Y);

        y = irk3(f,g,T,x,U,p); % Reduced Order
        l2o(a,b) = norm2(Y-y)/norm2(Y);

        b = b + 1;
        fprintf('#');
    end;

    a = a + 1;
    fprintf('\n');
end;

k2o
l2o

%% Output

save('l2.mat','k2g','k2o','l2g','l2o');

if(nargin==0), return; end
figure();
h = surf(l2);
xlim([1,16]);
ylim([1,16]);
set(gca,'ZScale','log');
set(gca,'XTick',1:2:16);
set(gca,'YTick',1:2:16);
set(gca,'XTickLabel',1:8:N);
set(gca,'YTickLabel',1:8:N);
ylabel('State Dimension')
xlabel('Parameter Dimension');
colormap(antijet);
set(h,'CData',log10(get(h,'CData')))
view(135,30);
if(o==1), print('-dsvg',[mfilename(),'.svg']); end;

%% ======== Integrator ========

function y = irk3(f,g,t,z,u,p)

    h = t(2);
    T = round(t(3)/h);

    k1 = h*f(z,u(:,1),p);
    k2 = h*f(z + 0.5*k1,u(:,1),p);
    k3r = h*f(z + 0.75*k2,u(:,1),p);
    z = z + (2.0/9.0)*k1 + (1.0/3.0)*k2 + (4.0/9.0)*k3r;

    y(:,1) = g(z,u(:,1),p);
    y(end,T) = 0;

    for t=2:T
        l1 = h*f(z,u(:,t),p);
        l2 = h*f(z + 0.5*l1,u(:,t),p);
        z = z + (2.0/3.0)*l1 + (1.0/3.0)*k1 + (5.0/6.0)*(l2 - k2);
        y(:,t) = g(z,u(:,t),p);
        k1 = l1;
        k2 = l2;
    end;

%% ======== Colormap ========

function m = antijet(n)
% antijet colormap
% by Christian Himpe 2014
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )

if(nargin<1 || isempty(n)), n = 256; end;
L = linspace(0,1,n);

R = -0.5*sin( L*(1.37*pi)+0.13*pi )+0.5;
G = -0.4*cos( L*(1.5*pi) )+0.4;
B = 0.3*sin( L*(2.11*pi) )+0.3;

m = [R;G;B]';

