function XP = optmor(f,g,s,t,r,q,nf,x0,ut,co,yd)
% optmor (Version 2.0.rc)
% by Christian Himpe, 2013-2015 ( http://wwwmath.uni-muenster.de/u/himpe )
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )
%
%
%
%
%
%
%
% + Initial Parameter Base: Random(0), Const(1), Prior(2)
% + State-space trajectory selection: POD-GREEDY(0), POD(1), PCA(2), MEAN(3)
% + Orthogonalization Algorithm: QR(0), SVD(0), ORTH(0), GS(0)
% + Add x States per Iteration
% + Optimization Algorithm: fminunc(0), fminsearch(1)
% + Optimization Norm: (P-Norm)
% + Lasso Regularization Weight: default(0)
% + Tikhonov Regularization Weight: default(0)
% + Data-Driven Regularization Weight: default(0)
% + Monte-Carlo Basis Enrichment (Base Size)
% + Random Seed: none(0)
% + Solver: IRK3(0), Custom Solver with handle(-1)
%*

    % Version Info
    if(nargin==1 && strcmp(f,'version')), XP = 2.0; return; end;

    % Default Arguments
    if(nargin<7 || isempty(nf)), nf = 0; end;
    if(nargin<8 || isempty(x0)), x0 = 0; end;
    if(nargin<9 || isempty(ut)), ut = 1; end;
    if(nargin<10 || isempty(co)), co = 1; end;
    if(nargin<11 || isempty(yd)), yd = 0; end;

    % System Constants
    J = s(1);
    N = s(2);
    O = s(3);
    h = t(2);
    T = round((t(3)-t(1))/h);
    Q = numel(q);

    a = nf(7);
    b = nf(8);
    c = nf(9);

%% LAZY ARGUMENTS

    if(numel(nf)==1), nf = zeros(1,12); end
    if(numel(x0)==1), x0 = x0*ones(N,1); end
    if(numel(ut)==1), ut = [ut*ones(J,1),sparse(J,T-1)]; end
    if(numel(co)==1), co = co*speye(Q); end

    %
    if(nf(4)==0), nf(4) = 1; end
    if(nf(6)==0), nf(6) = 2; end;

    % Precision Matrix
    W = ainv(co);

%% OBJECTIVE FUNCTIONAL ASSEMBLY

    switch(nf(6))
        case 1, % L1-norm optimization

            if(a==0 && b==0 && c==0) % None
                j = @(y,p) -norm1t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h); end
            if(a~=0 && b==0 && c==0) % L1
                j = @(y,p) -norm1t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W); end
            if(a==0 && b~=0 && c==0) % L2
                j = @(y,p) -norm1t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                 + b*norm2w(p,W); end
            if(a==0 && b==0 && c~=0) % DD
                j = @(y,p) -norm1t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                                 + c*norm2t(yd-y,h); end
            if(a~=0 && b~=0 && c==0) % L1 + L2
                j = @(y,p) -norm1t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W) + b*norm2w(p,W); end
            if(a~=0 && b==0 && c~=0) % L1 + DD
                j = @(y,p) -norm1t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W)                 + c*norm2t(yd-y,h); end
            if(a==0 && b~=0 && c~=0) % L2 + DD
                j = @(y,p) -norm1t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                 + b*norm2w(p,W) + c*norm2t(yd-y,h); end
            if(a~=0 && b~=0 && c~=0) % L1 + L2 + DD
                j = @(y,p) -norm1t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W) + b*norm2w(p,W) + c*norm2t(yd-y,h); end

        case 2, % L2-norm optimization

            if(a==0 && b==0 && c==0) % None
                j = @(y,p) -norm2t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h); end
            if(a~=0 && b==0 && c==0) % L1
                j = @(y,p) -norm2t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W); end
            if(a==0 && b~=0 && c==0) % L2
                j = @(y,p) -norm2t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                 + b*norm2w(p,W); end
            if(a==0 && b==0 && c~=0) % DD
                j = @(y,p) -norm2t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                                 + c*norm2t(yd-y,h); end
            if(a~=0 && b~=0 && c==0) % L1 + L2
                j = @(y,p) -norm2t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W) + b*norm2w(p,W); end
            if(a~=0 && b==0 && c~=0) % L1 + DD
                j = @(y,p) -norm2t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W)                 + c*norm2t(yd-y,h); end
            if(a==0 && b~=0 && c~=0) % L2 + DD
                j = @(y,p) -norm2t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                 + b*norm2w(p,W) + c*norm2t(yd-y,h); end
            if(a~=0 && b~=0 && c~=0) % L1 + L2 + DD
                j = @(y,p) -norm2t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W) + b*norm2w(p,W) + c*norm2t(yd-y,h); end

        case Inf, % L8-norm optimization

            if(a==0 && b==0 && c==0) % None
                j = @(y,p) -norm8t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h); end
            if(a~=0 && b==0 && c==0) % L1
                j = @(y,p) -norm8t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W); end
            if(a==0 && b~=0 && c==0) % L2
                j = @(y,p) -norm8t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                 + b*norm2w(p,W); end
            if(a==0 && b==0 && c~=0) % DD
                j = @(y,p) -norm8t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                                 + c*norm2t(yd-y,h); end
            if(a~=0 && b~=0 && c==0) % L1 + L2
                j = @(y,p) -norm8t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W) + b*norm2w(p,W); end
            if(a~=0 && b==0 && c~=0) % L1 + DD
                j = @(y,p) -norm8t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W)                 + c*norm2t(yd-y,h); end
            if(a==0 && b~=0 && c~=0) % L2 + DD
                j = @(y,p) -norm8t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h)                 + b*norm2w(p,W) + c*norm2t(yd-y,h); end
            if(a~=0 && b~=0 && c~=0) % L1 + L2 + DD
                j = @(y,p) -norm8t(ode(f,g,h,T,x0,ut,p,nf(12))-y,h) + a*norm1w(p,W) + b*norm2w(p,W) + c*norm2t(yd-y,h); end

    end


%% MAIN LOOP

    switch(nf(1))
        case 0, % Random
            p = rand(Q,1);
        case 1, % Constant
            p = ones(Q,1);
        case 2, % Prior
            p = q;
    end
p = q;
    z = ode(f,1,h,T,x0,ut,p,nf(12));
    if(nf(2)==0), nf2 = 1; else nf2 = nf(2); end;
    X = res(z,nf2,nf(4));

    P = p;

    for I=2:r

        if(nf(10))
            mc = orth(rand(Q,nf(10)));
            p = ones(nf(10),1);
            J = @(p) j(ode(@(x,u,p) X'*f(X*x,u,p),@(x,u,p) g(X*x,u,p),h,T,X'*x0,ut,P*(P'*(mc*p)),nf(12)),mc*p);
        else
            J = @(p) j(ode(@(x,u,p) X'*f(X*x,u,p),@(x,u,p) g(X*x,u,p),h,T,X'*x0,ut,P*(P'*p),nf(12)),p);
        end

        switch(nf(5))
            case 0, % Unconstrained
                p = fminunc(J,p,optimset('Display','off'));
            case 1, % Nelder-Mead
                p = fminsearch(J,p,optimset('Display','off'));
        end

        if(nf(10)), p = mc*p; end;
        z = ode(f,1,h,T,x0,ut,p,nf(12));
        if(nf(2)==0), z = z - X*(X'*z); end;
        x = res(z,nf(2),nf(4));

        P = incorp(P,p,nf(3));
        X = incorp(X,x,nf(3));
I;
    end

    XP = {X,P};
end


%% ======== Time Series Norms ========
function x = norm1t(X,h)

    x = h*sum(sum(abs(X)));
end

function x = norm2t(X,h)

    x = sqrt(h*sum(sum(X.*X)));
end

function x = norm8t(X,h)

    x = max(max(abs(X)));
end

function x = norm1w(X,w)

    x = sum(abs(w*X));
end

function x = norm2w(X,w)

    x = (X'*w')*(w*X);
end

function x = norm2(X)

     x = X'*X;
end

%% ======== STATE-SELECT ========
function v = res(u,o,s)

    switch(o)
        case 0, %
            v = pod1(u,s);
        case 1, %
            v = pod1(u,s);
        case 2, %
            v = pod1(bsxfun(@minus,u,mean(u,2),s));
        case 3, %
            v = mean(u,2);
    end;
end


%% ======== INCORPORATE ========
function v = incorp(U,u,o)

    switch(o)
        case 0, % QR
            [v,dummy] = qr([U,u],0);
        case 1, % SVD
            [v,dummy,dummy2] = svd([U,u],'econ');
        case 2, % ORTH
            v = orth([U,u]);
        case 3, % Gram-Schmidt
            % v = 
    end;
end

%% ======== FAST APPROXIMATE INVERSION ========
function x = ainv(m)

    d = 1.0./diag(m);
    n = numel(d);
    x = bsxfun(@times,m,-d);
    x = bsxfun(@times,x,d');
    x(1:n+1:end) = d;
end

%% ======== PRINCIPAL DIRECTION ========
function x = pod1(d,s)

    [x,D,V] = svds(d,s);
end

%% ======== ODE INTEGRATORS ========
function x = ode(f,g,h,T,z,u,p,O)

    if(O<0) % Custom Solver
        if(O==-2), global CUSTOM_ODE; end;

        if(isnumeric(g) && g==1) % Compute State Trajectory
            x = CUSTOM_ODE(f,h,T,z,u,p);
        else % Compute Output Trajectory
            v = CUSTOM_ODE(f,h,T,z,u,p);

            x(:,1) = g(v(:,1),u(:,1),p);
            x(end,T) = 0;

            for t=2:T
                x(:,t) = g(v(:,t),u(:,t),p);
            end;
        end;

    else % (Accelerated) Improved Runge-Kutta Method (IRK3)
        k1 = h*f(z,u(:,1),p);
        k2 = h*f(z + 0.5*k1,u(:,1),p);
        k3r = h*f(z + 0.75*k2,u(:,1),p);
        z = z + (2.0/9.0)*k1 + (1.0/3.0)*k2 + (4.0/9.0)*k3r; % Ralston (RK3)

        if(isnumeric(g) && g==1) % Compute State Trajectory
            x(:,1) = z;
            x(end,T) = 0;

            for t=2:T
                l1 = h*f(z,u(:,t),p);
                l2 = h*f(z + 0.5*l1,u(:,t),p);
                z = z + (2.0/3.0)*l1 + (1.0/3.0)*k1 + (5.0/6.0)*(l2 - k2);
                x(:,t) = z;
                k1 = l1;
                k2 = l2;
            end;
        else % Compute Output Trajectory
            x(:,1) = g(z,u(:,1),p);
            x(end,T) = 0;

            for t=2:T
                l1 = h*f(z,u(:,t),p);
                l2 = h*f(z + 0.5*l1,u(:,t),p);
                z = z + (2.0/3.0)*l1 + (1.0/3.0)*k1 + (5.0/6.0)*(l2 - k2);
                x(:,t) = g(z,u(:,t),p);
                k1 = l1;
                k2 = l2;
            end;
        end;
    end;
end

